"""Module containing GluePySparkContext class that extends GlueContext."""

from __future__ import annotations

from typing import TypedDict

from awsglue.context import GlueContext

from .documentdb import DocumentDBMixin
from .dynamodb import DynamoDBMixin
from .jdbc import JDBCMixin
from .kafka import KafkaMixin
from .kinesis import KinesisMixin
from .mongodb import MongoDBMixin
from .opensearch import OpenSearchMixin
from .s3 import S3Mixin


class GlueContextOptions(TypedDict, total=False):
    """Options to be passed as kwargs when instantiating a GlueContext object."""

    minPartitions: int
    targetPartitions: int


class GluePySparkContext(
    DocumentDBMixin,
    DynamoDBMixin,
    JDBCMixin,
    KafkaMixin,
    KinesisMixin,
    MongoDBMixin,
    OpenSearchMixin,
    S3Mixin,
    # awsglue has no typings yet
    GlueContext,  # type: ignore[misc]
):
    """A custom PySpark context that extends the functionality of the AWS GlueContext.

    This class combines the features of the mixin classes with the AWS
    GlueContext to provide additional functionality for working with
    JDBC and S3 data sources in a Glue job.
    """



from __future__ import annotations  # noqa: D100

from typing import TYPE_CHECKING

from glue_utils.pyspark.connection_types import ConnectionType

if TYPE_CHECKING:
    from awsglue.context import GlueContext
    from awsglue.dynamicframe import DynamicFrame

    from glue_utils.pyspark.connection_options import (
        JDBCConnectionOptions,
        RedshiftJDBCConnectionOptions,
    )


class SQLServerMixin:
    """Mixin for working with SQL Server connections."""

    def create_dynamic_frame_from_sqlserver(
        self: GlueContext,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from a SQL Server database connection.

        Parameters
        ----------
        connection_options : JDBCConnectionOptions
            The connection options for the SQL Server database.
        transformation_ctx : str, optional
            The name of the transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the SQL Server database.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.SQLSERVER.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_sqlserver(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to a SQL Server database using JDBC.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write to the SQL Server database.
        connection_options : JDBCConnectionOptions
            The JDBC connection options for the SQL Server database.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame that was written to the SQL Server database.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.SQLSERVER.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )


class MySQLMixin:
    """Mixin for working with MySQL connections."""

    def create_dynamic_frame_from_mysql(
        self: GlueContext,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from a MySQL database connection.

        Parameters
        ----------
        connection_options : JDBCConnectionOptions
            The connection options for the MySQL database.
        transformation_ctx : str, optional
            The transformation context for the DynamicFrame. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the MySQL database.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.MYSQL.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_mysql(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to a MySQL database using JDBC.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write to the MySQL database.
        connection_options : JDBCConnectionOptions
            The JDBC connection options for the MySQL database.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame after writing to the MySQL database.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.MYSQL.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )


class OracleMixin:
    """Mixin for working with Oracle connections."""

    def create_dynamic_frame_from_oracle(
        self: GlueContext,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from an Oracle database connection.

        Parameters
        ----------
        connection_options : JDBCConnectionOptions
            The connection options for connecting to the Oracle database.
        transformation_ctx : str, optional
            The name of the transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the Oracle database.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.ORACLE.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_oracle(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to an Oracle database using JDBC.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write to the Oracle database.
        connection_options : JDBCConnectionOptions
            The JDBC connection options for connecting to the Oracle database.
        transformation_ctx : str, optional
            The transformation context for the write operation. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame that was written to the Oracle database.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.ORACLE.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )


class PostgreSQLMixin:
    """Mixin for working with PostgreSQL connections."""

    def create_dynamic_frame_from_postgresql(
        self: GlueContext,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from a PostgreSQL database connection.

        Parameters
        ----------
        connection_options : JDBCConnectionOptions
            The connection options for the PostgreSQL database.
        transformation_ctx : str, optional
            The transformation context for the DynamicFrame. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the PostgreSQL database.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.POSTGRESQL.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_postgresql(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to a PostgreSQL database.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write to the database.
        connection_options : JDBCConnectionOptions
            The JDBC connection options for the database.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame that was written to the database.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.POSTGRESQL.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )


class RedshiftSQLMixin:
    """Mixin for working with Redshift connections."""

    def create_dynamic_frame_from_redshift(
        self: GlueContext,
        connection_options: RedshiftJDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from a Redshift database connection.

        Parameters
        ----------
        connection_options : RedshiftJDBCConnectionOptions
            The connection options for the Redshift database.
        transformation_ctx : str, optional
            The transformation context for the DynamicFrame. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the Redshift database.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.REDSHIFT.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_redshift(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: RedshiftJDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to a Redshift database.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write to the database.
        connection_options : RedshiftJDBCConnectionOptions
            The JDBC connection options for the database.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame that was written to the database.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.REDSHIFT.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )


class JDBCMixin(
    MySQLMixin,
    OracleMixin,
    PostgreSQLMixin,
    RedshiftSQLMixin,
    SQLServerMixin,
):
    """Mixin for working with JDBC connections."""



"""Module containing the Format enum."""

from enum import Enum


class Format(Enum):
    """Enum representing different file formats."""

    AVRO = "avro"
    CSV = "csv"
    JSON = "json"
    PARQUET = "parquet"
    ORC = "orc"
    XML = "xml"
    GROKLOG = "grokLog"
    ION = "ion"



from .connection_types import ConnectionType
from .context import GlueContextOptions, GluePySparkContext
from .formats import Format
from .job import GluePySparkJob

__all__ = [
    "ConnectionType",
    "Format",
    "GlueContextOptions",
    "GluePySparkContext",
    "GluePySparkJob",
]



from __future__ import annotations  # noqa: D100

from typing import TYPE_CHECKING

from glue_utils.pyspark.connection_types import ConnectionType

if TYPE_CHECKING:
    from awsglue.context import GlueContext
    from awsglue.dynamicframe import DynamicFrame

    from glue_utils.pyspark.connection_options import (
        MongoDBSinkConnectionOptions,
        MongoDBSourceConnectionOptions,
    )


class MongoDBMixin:
    """Mixin for working with Mongodb connections."""

    def create_dynamic_frame_from_mongodb(
        self: GlueContext,
        connection_options: MongoDBSourceConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from a Mongodb data source.

        Parameters
        ----------
        connection_options : MongodbSourceConnectionOptions
            The connection options for the Mongodb data source.
        transformation_ctx : str, optional
            The transformation context for the dynamic frame. Defaults to "".

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.MONGODB.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_mongodb(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: MongoDBSinkConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to Mongodb using the specified connection options.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write to Mongodb.
        connection_options : MongodbSinkConnectionOptions
            The connection options for Mongodb.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.MONGODB.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )



from __future__ import annotations  # noqa: D100

from typing import TYPE_CHECKING

from glue_utils.pyspark.connection_types import ConnectionType
from glue_utils.pyspark.formats import Format

if TYPE_CHECKING:
    from awsglue.context import GlueContext
    from awsglue.dynamicframe import DynamicFrame

    from glue_utils.pyspark.connection_options import (
        S3ParquetSourceConnectionOptions,
        S3SinkConnectionOptions,
        S3SourceConnectionOptions,
    )
    from glue_utils.pyspark.format_options import (
        CSVFormatOptions,
        JSONFormatOptions,
        ParquetFormatOptions,
        XMLFormatOptions,
    )


class S3CSVMixin:
    """Mixin for working with CSV files in S3."""

    def create_dynamic_frame_from_s3_csv(
        self: GlueContext,
        connection_options: S3SourceConnectionOptions,
        format_options: CSVFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from a CSV file stored in Amazon S3.

        Parameters
        ----------
        connection_options : S3SourceConnectionOptions
            The connection options for the S3 source.
        format_options : CSVFormatOptions | None, optional
            The format options for the CSV file. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the CSV file in S3.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.CSV.value,
            format_options=format_options or {},
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_s3_csv(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: S3SinkConnectionOptions,
        format_options: CSVFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to a CSV file in Amazon S3.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write.
        connection_options : S3SinkConnectionOptions
            The connection options for the S3 sink.
        format_options : CSVFormatOptions | None, optional
            The format options for the CSV file. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The written DynamicFrame.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.CSV.value,
            format_options=format_options or {},
            transformation_ctx=transformation_ctx,
        )


class S3ParquetMixin:
    """Mixin for working with Parquet files in S3."""

    def create_dynamic_frame_from_s3_parquet(
        self: GlueContext,
        connection_options: S3ParquetSourceConnectionOptions,
        format_options: ParquetFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from Parquet files stored in Amazon S3.

        Parameters
        ----------
        connection_options : S3SourceConnectionOptions
            The connection options for the S3 source.
        format_options : ParquetFormatOptions | None, optional
            The format options for reading Parquet files. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the Parquet files in S3.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.PARQUET.value,
            format_options=format_options or {},
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_s3_parquet(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: S3SinkConnectionOptions,
        format_options: ParquetFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to Parquet format in Amazon S3.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write.
        connection_options : S3SinkConnectionOptions
            The connection options for writing to S3.
        format_options : ParquetFormatOptions | None, optional
            The format options for writing Parquet. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The written DynamicFrame.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.PARQUET.value,
            format_options=format_options or {},
            transformation_ctx=transformation_ctx,
        )


class S3JSONMixin:
    """Mixin for working with JSON files in S3."""

    def create_dynamic_frame_from_s3_json(
        self: GlueContext,
        connection_options: S3SourceConnectionOptions,
        format_options: JSONFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from JSON data stored in Amazon S3.

        Parameters
        ----------
        connection_options : S3SourceConnectionOptions
            The connection options for the S3 source.
        format_options : JSONFormatOptions | None, optional
            The format options for the JSON data. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the JSON data in S3.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.JSON.value,
            format_options=format_options or {},
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_s3_json(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: S3SinkConnectionOptions,
        format_options: JSONFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to JSON format and save it in S3.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to be written.
        connection_options : S3SinkConnectionOptions
            The connection options for S3 sink.
        format_options : JSONFormatOptions | None, optional
            The format options for JSON. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame after writing to S3.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.JSON.value,
            format_options=format_options or {},
            transformation_ctx=transformation_ctx,
        )


class S3XMLMixin:
    """Mixin for working with XML files in S3."""

    def create_dynamic_frame_from_s3_xml(
        self: GlueContext,
        connection_options: S3SourceConnectionOptions,
        format_options: XMLFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from XML data stored in Amazon S3.

        Parameters
        ----------
        connection_options : S3SourceConnectionOptions
            The connection options for the S3 data source.
        format_options : XMLFormatOptions, optional
            The format options for parsing the XML data. Defaults to None.
        transformation_ctx : str, optional
            The transformation context for the DynamicFrame. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame containing the parsed XML data.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.XML.value,
            format_options=format_options or {},
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_s3_xml(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: S3SinkConnectionOptions,
        format_options: XMLFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to XML format and saves it in S3.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to be written.
        connection_options : S3SinkConnectionOptions
            The connection options for S3 sink.
        format_options : XMLFormatOptions | None, optional
            The format options for XML. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame after writing to S3.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.XML.value,
            format_options=format_options or {},
            transformation_ctx=transformation_ctx,
        )


class S3Mixin(S3CSVMixin, S3ParquetMixin, S3JSONMixin, S3XMLMixin):
    """Mixin for working with S3 connections."""



"""Module providing the GluePySparkJob class for handling Glue ETL jobs."""

from __future__ import annotations

import sys
from contextlib import contextmanager
from dataclasses import fields
from enum import Enum
from typing import TYPE_CHECKING, Generic, cast, overload

from awsglue.job import Job
from awsglue.utils import getResolvedOptions
from pyspark import SparkConf, SparkContext
from typing_extensions import TypeVar

from glue_utils import BaseOptions

from .context import GlueContextOptions, GluePySparkContext

if TYPE_CHECKING:
    from collections.abc import Generator

    from pyspark.sql import SparkSession

T = TypeVar("T", bound=BaseOptions, default=BaseOptions)


class GluePySparkJob(Generic[T]):
    """Class that handles the boilerplate setup for Glue ETL jobs."""

    class LogLevel(Enum):
        """Enum class representing different log levels.

        Attributes
        ----------
            ALL (str): The lowest log level, includes all log messages.
            DEBUG (str): Log level for debugging information.
            ERROR (str): Log level for error messages.
            FATAL (str): Log level for fatal error messages.
            INFO (str): Log level for informational messages.
            OFF (str): The highest log level, turns off all log messages.
            TRACE (str): Log level for detailed trace messages.
            WARN (str): Log level for warning messages.

        """

        ALL = "ALL"
        DEBUG = "DEBUG"
        ERROR = "ERROR"
        FATAL = "FATAL"
        INFO = "INFO"
        OFF = "OFF"
        TRACE = "TRACE"
        WARN = "WARN"

    options: T
    sc: SparkContext
    spark: SparkSession
    glue_context: GluePySparkContext

    @overload
    def __init__(
        self: GluePySparkJob[BaseOptions],
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[BaseOptions],
        *,
        spark_conf: SparkConf,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[BaseOptions],
        *,
        glue_context_options: GlueContextOptions,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[BaseOptions],
        *,
        log_level: LogLevel,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[BaseOptions],
        *,
        spark_conf: SparkConf,
        glue_context_options: GlueContextOptions,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[BaseOptions],
        *,
        spark_conf: SparkConf,
        log_level: LogLevel,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[BaseOptions],
        *,
        glue_context_options: GlueContextOptions,
        log_level: LogLevel,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[BaseOptions],
        *,
        spark_conf: SparkConf,
        glue_context_options: GlueContextOptions,
        log_level: LogLevel,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[T],
        *,
        options_cls: type[T],
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[T],
        *,
        options_cls: type[T],
        spark_conf: SparkConf,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[T],
        *,
        options_cls: type[T],
        glue_context_options: GlueContextOptions,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[T],
        *,
        options_cls: type[T],
        log_level: LogLevel,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[T],
        *,
        options_cls: type[T],
        spark_conf: SparkConf,
        glue_context_options: GlueContextOptions,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[T],
        *,
        options_cls: type[T],
        spark_conf: SparkConf,
        log_level: LogLevel,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[T],
        *,
        options_cls: type[T],
        glue_context_options: GlueContextOptions,
        log_level: LogLevel,
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[T],
        *,
        options_cls: type[T],
        spark_conf: SparkConf,
        glue_context_options: GlueContextOptions,
        log_level: LogLevel,
    ) -> None: ...

    def __init__(
        self,
        *,
        options_cls: type[T | BaseOptions] = BaseOptions,
        spark_conf: SparkConf | None = None,
        glue_context_options: GlueContextOptions | None = None,
        log_level: LogLevel = LogLevel.WARN,
    ) -> None:
        """Initialize a Job object.

        Parameters
        ----------
        options_cls : type[T | BaseOptions], optional
            The class representing the options for the job. Defaults to
            BaseOptions.
        spark_conf : SparkConf | None, optional
            The Spark configuration. Defaults to None.
        glue_context_options : GlueContextOptions | None, optional
            The Glue context options. Defaults to None.
        log_level : LogLevel, optional
            The log level for the job. Defaults to LogLevel.WARN.

        """
        if not issubclass(options_cls, BaseOptions):
            msg = "options_cls must be a subclass of BaseOptions."
            raise TypeError(msg)

        field_names = {field.name for field in fields(options_cls)}

        params = []
        if "--JOB_NAME" in sys.argv:
            params.append("JOB_NAME")
        params.extend(field_names)

        glue_args = getResolvedOptions(sys.argv, params)

        self.options = cast(T, options_cls.from_options(glue_args))

        self.sc = self.create_spark_context(spark_conf)
        self.set_log_level(log_level)

        self.glue_context = self.create_glue_context(glue_context_options)
        self.spark = self.glue_context.spark_session

        self._job = Job(self.glue_context)
        self._job.init(glue_args.get("JOB_NAME", ""), glue_args)

    def create_spark_context(self, conf: SparkConf | None = None) -> SparkContext:
        """Create a SparkContext.

        Parameters
        ----------
        conf, optional
            The SparkConf to use, by default None

        Returns
        -------
        SparkContext
            The SparkContext created.

        """
        if conf:
            if not isinstance(conf, SparkConf):
                msg = "conf must be an instance of SparkConf."
                raise TypeError(msg)
            return SparkContext.getOrCreate(conf=conf)
        return SparkContext.getOrCreate()

    def create_glue_context(
        self,
        glue_context_options: GlueContextOptions | None = None,
    ) -> GluePySparkContext:
        """Create a GluePySparkContext object using the provided SparkContext and PartitionOptions.

        Parameters
        ----------
        glue_context_options : GlueContextOptions | None, optional
            Optional PartitionOptions object containing additional
            options for configuring the GlueContext.

        """
        if glue_context_options:
            return GluePySparkContext(self.sc, **glue_context_options)
        return GluePySparkContext(self.sc)

    @contextmanager
    def managed_glue_context(
        self,
        *,
        commit: bool = True,
    ) -> Generator[GluePySparkContext, None, None]:
        """Context manager for managing the GluePySparkContext.

        Parameters
        ----------
        commit, optional
            Whether to commit the job, by default True

        """
        yield self.glue_context
        if commit:
            self.commit()

    def commit(self) -> None:
        """Commit the Glue ETL job."""
        self._job.commit()

    def set_log_level(
        self,
        level: LogLevel,
    ) -> None:
        """Set the log level for the SparkContext.

        Parameters
        ----------
        level : LogLevel
            The log level to be set.

        """
        self.sc.setLogLevel(level.value)



"""Module containing small-utility functions that can be useful when working with Glue jobs."""

from __future__ import annotations

from typing import Any


def generate_partitioned_path(**kwargs: Any) -> str:  # noqa: ANN401
    """Generate a partitioned path by joining key-value pairs with separators.

    Parameters
    ----------
    **kwargs : Any
        Key-value pairs representing the partitions.

    Returns
    -------
    str
        The generated partitioned path.

    Examples
    --------
    >>> generate_partitioned_path(year=2022, month=10, day=15)
    'year=2022/month=10/day=15'
    >>> generate_partitioned_path(category='electronics', brand='apple')
    'category=electronics/brand=apple'

    """
    return "/".join(f"{key}={value}" for key, value in kwargs.items())



"""Module containing dictionary structures for handling different connection types."""

from typing import Literal, TypedDict


class BookmarkConnectionOptions(TypedDict, total=False):
    """TypedDict class representing general connection options.

    Reference
    ---------
    - https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect.html#aws-glue-programming-etl-connect-general-options

    """

    jobBookmarkKeys: list[str]
    jobBookmarkKeysSortOrder: Literal["asc", "desc"]


class JDBCConnectionOptions(BookmarkConnectionOptions, total=False):
    """Connection options for JDBC connections.

    Reference
    ---------
    - https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-jdbc-home.html

    """

    useConnectionProperties: Literal["true"]
    connectionName: str
    databaseName: str
    url: str
    dbtable: str
    user: str
    password: str
    customJdbcDriverS3Path: str
    customJdbcDriverClassName: str
    bulkSize: str
    hashfield: str
    hashexpression: str
    hashpartitions: str
    sampleQuery: str
    enablePartitioningForSampleQuery: Literal["true", "false"]
    sampleSize: str


# Redshift has a connection option with a dot in it, so we need to use
# this form of TypedDict to define the connection options.
RedshiftOptions = TypedDict(
    "RedshiftOptions",
    {
        "autopushdown": Literal["true", "false"],
        "autopushdown.s3_result_cache": Literal["true", "false"],
        "aws_iam_role": str,
        "csvnullstring": str,
        "DbUser": str,
        "extracopyoptions": str,
        "redshiftTmpDir": str,
        "sse_kms_key": str,
        "unload_s3_format": Literal["TEXT", "PARQUET"],
    },
    total=False,
)


class RedshiftJDBCConnectionOptions(
    RedshiftOptions, JDBCConnectionOptions, total=False
):
    """Connection options for Redshift JDBC connections.

    Reference
    ---------
    - https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-redshift-home.html

    """


class S3SourceConnectionOptions(BookmarkConnectionOptions, total=False):
    """Connection options for S3 connections.

    Reference
    ---------
    - AWS Glue Programming ETL Connect to S3:
      https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-s3-home.html

    """

    paths: list[str]
    exclusions: str
    compressionType: Literal["gzip", "bzip2"]
    groupFiles: Literal["inPartition", "none"]
    groupSize: str
    recurse: bool
    maxBand: str
    maxFilesInBand: str
    isFailFast: bool
    catalogPartitionPredicate: str
    excludeStorageClasses: list[str]
    useS3ListImplementation: bool


class S3ParquetSourceConnectionOptions(S3SourceConnectionOptions, total=False):
    """Additional connection options when using Parquet files in S3.

    Reference
    ---------
    - Using the Parquet format in AWS Glue:
      https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-format-parquet-home.html

    """

    mergeSchema: bool


class S3SinkConnectionOptions(BookmarkConnectionOptions, total=False):
    """Connection options for writing data to an S3 sink."""

    path: str
    compression: Literal["gzip", "bzip2"]
    partitionKeys: list[str]


# https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-dynamodb-home.html
DynamoDBSourceConnectionOptions = TypedDict(
    "DynamoDBSourceConnectionOptions",
    {
        "dynamodb.input.tableName": str,
        "dynamodb.throughput.read.percent": str,
        "dynamodb.splits": str,
        "dynamodb.sts.roleArn": str,
        "dynamodb.sts.roleSessionName": str,
    },
    total=False,
)

# https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-dynamodb-home.html
DynamoDBExportSourceConnectionOptions = TypedDict(
    "DynamoDBExportSourceConnectionOptions",
    {
        "dynamodb.export": Literal["ddb", "s3"],
        "dynamodb.tableArn": str,
        "dynamodb.s3.bucket": str,
        "dynamodb.s3.prefix": str,
        "dynamodb.s3.bucketOwner": str,
        "dynamodb.simplifyDDBJson": Literal["true", "false"],
        "dynamodb.exportTime": str,
        "dynamodb.sts.region": str,
    },
    total=False,
)

# https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-dynamodb-home.html
DynamoDBSinkConnectionOptions = TypedDict(
    "DynamoDBSinkConnectionOptions",
    {
        "dynamodb.output.tableName": str,
        "dynamodb.throughput.write.percent": str,
        "dynamodb.output.numParallelTasks": str,
        "dynamodb.output.retry": str,
        "dynamodb.sts.roleArn": str,
        "dynamodb.sts.roleSessionName": str,
    },
    total=False,
)


class KinesisConnectionOptions(TypedDict, total=False):
    """Connection options for Kinesis connections.

    Reference
    ---------
    - https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-kinesis-home.html

    """

    streamARN: str
    failOnDataLoss: Literal["true", "false"]
    awsSTSRoleArn: str
    awsSTSSessionName: str
    awsSTSEndpoint: str


class KinesisSourceConnectionOptions(KinesisConnectionOptions, total=False):
    """Connection options to read from Kinesis."""

    classification: str
    streamName: str
    endpointUrl: str
    delimiter: str
    startingPosition: str
    maxFetchTimeInMs: str
    maxFetchRecordsPerShard: str
    maxRecordPerRead: str
    addIdleTimeBetweenReads: Literal["true", "false"]
    idleTimeBetweenReadsInMs: str
    describeShardInterval: str
    numRetries: str
    retryIntervalMs: str
    maxRetryIntervalMs: str
    avoidEmptyBatches: Literal["true", "false"]
    schema: str
    inferSchema: Literal["true", "false"]
    addRecordTimestamp: Literal["true", "false"]
    emitConsumerLagMetrics: Literal["true", "false"]
    fanoutConsumerARN: str


class KinesisSinkConnectionOptions(KinesisConnectionOptions, total=False):
    """Connection options for writing to Kinesis."""

    partitionKey: str
    recordMaxBufferedTime: str
    aggregationEnabled: Literal["true", "false"]
    aggregationMaxSize: str
    aggregationMaxCount: str
    producerRateLimit: str
    collectionMaxCount: str
    collectionMaxSize: str


# https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-documentdb-home.html
DocumentDBSourceConnectionOptions = TypedDict(
    "DocumentDBSourceConnectionOptions",
    {
        "uri": str,
        "database": str,
        "collection": str,
        "username": str,
        "password": str,
        "ssl": str,
        "ssl.domain_match": str,
        "batchSize": str,
        "partitioner": str,
        "partitionerOptions.partitionKey": str,
        "partitionerOptions.partitionSizeMB": str,
        "partitionerOptions.samplesPerPartition": str,
        "partitionerOptions.shardKey": str,
        "partitionerOptions.numberOfPartitions": str,
    },
    total=False,
)


class DocumentDBSinkConnectionOptions(TypedDict, total=False):
    """Connection options for writing to DocumentDB.

    Reference:
    - https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-documentdb-home.html
    """

    uri: str
    database: str
    collection: str
    username: str
    password: str
    extendedBSONTypes: str
    replaceDocument: str
    maxBatchSize: str
    retryWrites: str


# https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-mongodb-home.html
MongoDBSourceConnectionOptions = TypedDict(
    "MongoDBSourceConnectionOptions",
    {
        "connectionName": str,
        "uri": str,
        "connection.uri": str,
        "username": str,
        "password": str,
        "database": str,
        "collection": str,
        "ssl": Literal["true", "false"],
        "ssl.domain_match": Literal["true", "false"],
        "disableUpdateUri": Literal["true", "false"],
        "batchSize": str,
        "partitioner": str,
        "partitionerOptions.partitionKey": str,
        "partitionerOptions.partitionSizeMB": str,
        "partitionerOptions.samplesPerPartition": str,
        "partitionerOptions.shardKey": str,
        "partitionerOptions.numberOfPartitions": str,
    },
    total=False,
)

# https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-mongodb-home.html
MongoDBSinkConnectionOptions = TypedDict(
    "MongoDBSinkConnectionOptions",
    {
        "connectionName": str,
        "uri": str,
        "connection.uri": str,
        "username": str,
        "password": str,
        "database": str,
        "collection": str,
        "ssl": Literal["true", "false"],
        "ssl.domain_match": Literal["true", "false"],
        "disableUpdateUri": Literal["true", "false"],
        "extendedBSONTypes": Literal["true", "false"],
        "replaceDocument": Literal["true", "false"],
        "maxBatchSize": str,
        "retryWrites": Literal["true", "false"],
    },
    total=False,
)

# https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-opensearch-home.html
OpenSearchSourceConnectionOptions = TypedDict(
    "OpenSearchSourceConnectionOptions",
    {
        "connectionName": str,
        "opensearch.resource": str,
        "opensearch.query": str,
        "pushdown": Literal["true", "false"],
        "opensearch.read.field.as.array.include": str,
    },
    total=False,
)

# https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-opensearch-home.html
OpenSearchSinkConnectionOptions = TypedDict(
    "OpenSearchSinkConnectionOptions",
    {
        "connectionName": str,
        "opensearch.resource": str,
    },
    total=False,
)

# https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-kafka-home.html
KafkaSourceConnectionOptions = TypedDict(
    "KafkaSourceConnectionOptions",
    {
        "connectionName": str,
        "bootstrap.servers": str,
        "security.protocol": Literal["SSL", "PLAINTEXT"],
        "topicName": str,
        "assign": str,
        "subscribePattern": str,
        "classification": str,
        "delimiter": str,
        "startingOffsets": Literal["earliest", "latest"],
        "startingTimestamp": str,
        "endingOffsets": str,
        "pollTimeoutMs": str,
        "numRetries": str,
        "retryIntervalMs": str,
        "maxOffsetsPerTrigger": str,
        "minPartitions": str,
        "includeHeaders": Literal["true", "false"],
        "schema": str,
        "inferSchema": Literal["true", "false"],
        "addRecordTimestamp": Literal["true", "false"],
        "emitConsumerLagMetrics": Literal["true", "false"],
    },
    total=False,
)


class KafkaSinkConnectionOptions(TypedDict, total=False):
    """Connection options for writing to Kafka.

    Reference:
    - https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-kafka-home.html
    """

    connectionName: str
    topicName: str
    partition: str
    key: str
    classification: str



"""Module containing dictionary structures for handling different format options."""

from __future__ import annotations

from typing import Literal, TypedDict


class S3FormatOptions(TypedDict, total=False):
    """Common format options for S3."""

    attachFilename: str
    attachTimestamp: str


class CSVFormatOptions(S3FormatOptions, total=False):
    """Format options for CSV files.

    Reference
    ---------
    - Using the CSV format in AWS Glue:
      https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-format-csv-home.html

    """

    separator: str
    escaper: str
    quoteChar: str | Literal[-1]
    multiLine: bool
    withHeader: bool
    writeHeader: bool
    skipFirst: bool
    optimizePerformance: bool
    strictCheckForQuoting: bool


class ParquetFormatOptions(S3FormatOptions, total=False):
    """Format options for Parquet files.

    Reference
    ---------
    - Using the Parquet format in AWS Glue:
      https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-format-parquet-home.html

    """

    useGlueParquetWriter: bool
    compression: Literal[
        "uncompressed",
        "snappy",
        "gzip",
        "lzo",
    ]
    blockSize: int
    pageSize: int


class JSONFormatOptions(S3FormatOptions, total=False):
    """Format options for JSON files.

    Reference
    ---------
    - Using the JSON format in AWS Glue:
      https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-format-json-home.html

    """

    jsonPath: str
    multiline: bool
    optimizePerformance: bool
    withSchema: str


class XMLFormatOptions(S3FormatOptions, total=False):
    """Format options for XML files.

    Reference
    ---------
    - Using the XML format in AWS Glue:
      https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-format-xml-home.html

    """

    rowTag: str
    encoding: str
    excludeAttribute: bool
    treatEmptyValueAsNull: bool
    attributePrefix: str
    valueTag: str
    ignoreSurroundingSpaces: bool
    withSchema: str
